"""
Define a suite a tests for the Component module.
"""
import pytest

from striplog.rock import Rock
from striplog import Component
from striplog import Lexicon

r = {'colour': 'grey',
     'grainsize': 'vf-f',
     'lithology': 'sand'}

r2 = {'grainsize': 'VF-F',
      'colour': 'Grey',
      'lithology': 'Sand'}

r3 = {'grainsize': 'Coarse',
      'colour': 'Grey',
      'lithology': 'Sandstone'}

r6 = {'grainsize': 'Coarse',
      'colour': 'Grey',
      'lithology': None}

r7 = {'coords': (123, 456)}

r8 = {'coords': (123, 456),
      'None value': None}

def test_init():
    """Test Rock() for backward compatibility.
    """
    with pytest.warns(DeprecationWarning):
        rock = Rock(r)
    assert rock

    rock = Component(r)
    assert rock.colour == 'grey'


def test_component_html():
    """
    For jupyter notebook.

    Hard to test this because attributes are returned in random order.
    """
    r = {'lithology': 'sand'}
    rock = Component(r)
    html = """<table><tr><td><strong>lithology</strong></td><td>sand</td></tr></table>"""
    assert rock._repr_html_() == html


def test_identity():
    """
    Test equals.
    """
    rock = Component(r)
    assert rock != 'non-Component'

    rock2 = Component(r2)
    assert rock == rock2

    rock3 = Component(r3)
    assert rock != rock3


def test_summary():
    """
    Test ability to generate summaries.
    """
    rock = Component(r)
    s = rock.summary(fmt="My rock: {lithology} ({colour}, {grainsize!u})")
    assert s == 'My rock: sand (grey, VF-F)'

    rock6 = Component(r6)
    s = rock6.summary(fmt="My rock: {lithology}")
    assert s == 'My rock: _'

    empty = Component({})
    d = "String"
    assert not empty  # Should have False value
    assert empty.summary(default=d) == d


def test_from_text():
    """
    Test generation from strings.
    """
    rock3 = Component(r3)
    lexicon = Lexicon.default()
    s = 'Grey coarse sandstone.'
    rock4 = Component.from_text(s, lexicon)
    assert rock3 == rock4
    rock5 = Component.from_text(s, lexicon, required='not there')
    assert not rock5  # Should be None

def test_None():
    """
    Test generation with None
    """
    rock8 = Component(r8)
    assert rock8 == Component({'coords': (123, 456)})

def test_tuple():
    """
    Test generation from tuples
    """
    rock7 = Component(r7)
    assert rock7 == Component({'coords': (123, 456)})



# striplog documentation build configuration file.
#
import sys
import os

sys.path.insert(0, os.path.abspath('..'))


# -- Setup function ----------------------------------------------------------

# Defines custom steps in the process.

def autodoc_skip_member(app, what, name, obj, skip, options):
    """Exclude all private attributes, methods, and dunder methods from Sphinx."""
    import re
    exclude = re.findall(r'\._.*', str(obj))
    return skip or exclude

def remove_module_docstring(app, what, name, obj, options, lines):
    """Remove everything after 'Author: '."""
    if what == "module":
        keep = [i for i, line in enumerate(lines) if line.startswith("Author: ")]
        if keep:
            del lines[keep[0]:]
    return

def setup(app):
    app.connect('autodoc-skip-member', autodoc_skip_member)
    app.connect("autodoc-process-docstring", remove_module_docstring)
    return


# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinxcontrib.apidoc',
    'sphinx.ext.githubpages',
    'sphinx.ext.coverage',
    'sphinx.ext.napoleon',
    'myst_nb'
]

myst_enable_extensions = ["dollarmath", "amsmath"]

# Apidoc automation
# https://pypi.org/project/sphinxcontrib-apidoc/
# The apidoc extension and this code automatically update apidoc.
apidoc_module_dir = '../striplog'
apidoc_output_dir = './api'
apidoc_excluded_paths = []
apidoc_toc_file = False
apidoc_separate_modules = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# General information about the project.
project = 'striplog'
copyright = '2022, The Striplog Authors'
author = 'The Striplog Authors'

# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build',]


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'furo'

html_theme_options = {
    "sidebar_hide_name": True,
}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/striplog_logo.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = [
    'custom.css',
]



"""
Run all the tests.

This is the same as doing this on the command line:

  py.test --mpl --cov striplog

We have to run tests this way because we need to set the
matplotlib backend for travis.
"""
import pytest
import matplotlib
matplotlib.use('agg')

# To generate new images, uncomment this line and instead
# comment out the last one (below).
# pytest.main(["--mpl-generate-path=tests/baseline", "--cov", "striplog"])

# To test as normal
pytest.main(["--mpl", "--cov", "striplog"])



"""
Defines a suite a tests for images run with:

    py.test --mpl

To generate new test images, see the instructions in
striplog/run_tests.py

https://pypi.python.org/pypi/pytest-mpl/0.3
"""
import random

import matplotlib.pyplot as plt
import pytest

from striplog import Striplog, Lexicon
from striplog import Legend, Component, Decor
from striplog.markov import Markov_chain

from .test_striplog import las3


params = {'tolerance': 20,
          'savefig_kwargs': {'dpi': 100},
          }


@pytest.mark.mpl_image_compare(**params)
def test_striplog_plot():
    """
    Tests mpl image of striplog.
    """
    legend = Legend.builtin('NSDOE')

    imgfile = "tests/data/M-MG-70_14.3_135.9.png"

    striplog = Striplog.from_image(imgfile, 14.3, 135.9, legend=legend)

    fig = striplog.thickest(n=5).plot(legend=legend,
                                      ladder=False,
                                      return_fig=True)
    return fig


@pytest.mark.mpl_image_compare(**params)
def test_striplog_ladder_plot():
    """
    Tests mpl image of striplog with the ladder option.
    """
    legend = Legend.builtin('NSDOE')

    imgfile = "tests/data/M-MG-70_14.3_135.9.png"

    striplog = Striplog.from_image(imgfile, 14.3, 135.9, legend=legend)

    fig = striplog.thickest(n=5).plot(legend=legend,
                                      ladder=True,
                                      return_fig=True)
    return fig


@pytest.mark.mpl_image_compare(**params)
def test_striplog_colour_plot():
    """
    Tests mpl image of striplog with the ladder option.
    """
    legend = Legend.builtin('NSDOE')

    imgfile = "tests/data/M-MG-70_14.3_135.9.png"

    striplog = Striplog.from_image(imgfile, 14.3, 135.9, legend=legend)

    for iv in striplog:
        iv.data['porosity'] = iv.top.z/100

    fig = striplog.plot(colour='porosity', aspect=3, return_fig=True)

    return fig


@pytest.mark.mpl_image_compare(**params)
def test_striplog_point_plot():
    """
    Tests mpl image of striplog with the points option.
    """
    legend = Legend.builtin('NSDOE')

    imgfile = "tests/data/M-MG-70_14.3_135.9.png"

    striplog = Striplog.from_image(imgfile, 14.3, 135.9, legend=legend)

    for iv in striplog:
        iv.data['porosity'] = iv.top.z/100

    fig = striplog.plot(style='points',
                        field='porosity',
                        aspect=3,
                        return_fig=True)

    return fig


@pytest.mark.mpl_image_compare(**params)
def test_striplog_top_plot():
    """
    Tests mpl image of striplog with the tops option.
    """
    tops_csv = """top, Comp formation
                   25, Escanilla Fm.
                   35, San Vicente Fm.
                   20, Sobrarbe Fm.
                   50, Cretaceous"""

    tops = Striplog.from_csv(text=tops_csv)

    fig = tops.plot(style='tops',
                    field='formation',
                    aspect=1.5,
                    return_fig=True)

    return fig


@pytest.mark.mpl_image_compare(**params)
def test_decor_plot():
    """
    Tests mpl image of decor.
    """
    r = {'colour': 'grey',
         'grainsize': 'vf-f',
         'lithology': 'sand'}

    rock = Component(r)

    d = {'color': '#86F0B6',
         'component': rock,
         'width': 3}

    decor = Decor(d)

    fig = plt.figure(figsize=(4, 1))
    fig = decor.plot(fmt="{lithology!t} {colour} {grainsize}", fig=fig)
    return fig


@pytest.mark.mpl_image_compare(**params)
def test_pattern_fills():
    """
    Tests mpl image of decors with pattern fills.
    """
    hatches = "pctLbs!=v^"
    decors = [Decor({'component': Component({'hatch': h}),
                     'hatch': h,
                     'colour': '#eeeeee'})
              for h in hatches]

    fig = plt.figure(figsize=(1, 12))
    for i, d in enumerate(decors):
        ax = fig.add_subplot(len(decors), 1, i+1)
        ax = d.plot(ax=ax, fmt='')
    return fig


@pytest.mark.mpl_image_compare(**params)
def test_histogram():
    """Test histogram plot.
    """
    fig, ax = plt.subplots()
    lexicon = Lexicon.default()
    striplog = Striplog.from_las3(las3, lexicon=lexicon)
    *_, ax = striplog.histogram(ax=ax)
    return fig


@pytest.mark.mpl_image_compare(**params)
def test_bar():
    """Test bar plot.
    """
    fig, ax = plt.subplots()
    lexicon = Lexicon.default()
    striplog = Striplog.from_las3(las3, lexicon=lexicon)
    legend = Legend.builtin('nagmdm__6_2')
    ax = striplog.bar(sort=True, legend=legend, ax=ax, align='center')
    return fig


@pytest.mark.mpl_image_compare(**params)
def test_markov():
    """Test Markov plot.
    """
    fig, ax = plt.subplots()
    data = "sssmmmlllmlmlsslsllsmmllllmssssllllssmmlllllssssssmmmmsmllllssslmslmsmmmslsllll"""
    m = Markov_chain.from_sequence(data, include_self=True)
    ax = m.plot_norm_diff(ax=ax)
    return fig


@pytest.mark.mpl_image_compare(**params)
def test_markov_graph_plot():
    """Test Markov plot.
    """
    fig, ax = plt.subplots()
    data = "sssmmmlllmlmlsslsllsmmllllmssssllllssmmlllllssssssmmmmsmllllssslmslmsmmmslsllll"""
    m = Markov_chain.from_sequence(data, include_self=False)
    ax = m.plot_graph(ax=ax, seed=42)
    return fig




import sys
import glob
import json
import pathlib
import shutil


def change_kernel(notebook):
    """
    Vanillafy the kernelspec.
    """
    new_kernelspec = {
        "display_name": "Python 3 (ipykernel)",
        "language": "python",
        "name": "python3",
    }
    notebook['metadata']['kernelspec'].update(new_kernelspec)
    return notebook


def main(path):
    """
    Process the IPYNB files in path, save in place (side-effect).
    """
    fnames = glob.glob(path.strip('/') + '/[!_]*.ipynb')  # Not files with underscore.

    for fname in fnames:
        with open(fname, encoding='utf-8') as f:
            notebook = json.loads(f.read())

        new_nb = change_kernel(notebook)

        with open(fname, 'w') as f:
            _ = f.write(json.dumps(new_nb))

    return


if __name__ == '__main__':
    print(sys.argv[1])
    _ = main(sys.argv[1])


"""
Define a suite a tests for the Striplog module.
"""
import numpy as np
import pytest

from striplog import Component
from striplog import Interval
from striplog import Legend
from striplog import Lexicon
from striplog import Striplog
from striplog.striplog import StriplogError

las3 = """~Lithology_Parameter
LITH .   Striplog         : Lithology source          {S}
LITHD.   MD               : Lithology depth reference {S}

~Lithology_Definition
LITHT.M                   : Lithology top depth       {F}
LITHB.M                   : Lithology base depth      {F}
LITHD.                    : Lithology description     {S}

~Lithology_Data | Lithology_Definition
  280.000,  299.986,  "Red, siltstone"
  299.986,  304.008,  "Grey, sandstone, vf-f"
  304.008,  328.016,  "Red, siltstone"
  328.016,  328.990,  "Limestone"
  328.990,  330.007,  "Red, siltstone"
  330.007,  333.987,  "Limestone"
  333.987,  338.983,  "Red, siltstone"
  338.983,  340.931,  "Limestone"
  340.931,  345.927,  "Red, siltstone"
  345.927,  346.944,  "Limestone"
  346.944,  414.946,  "Red, siltstone"
  414.946,  416.936,  "Grey, mudstone"
  416.936,  422.440,  "Red, heterolithic"
  422.440,  423.414,  "Grey, mudstone"
 """

csv_intervals = """   200.000,  230.329,  Anhydrite
                      230.329,  233.269,  Grey vf-f sandstone
                      233.269,  234.700,  Anhydrite
                      234.700,  236.596,  Dolomite
                      236.596,  237.911,  Red siltstone
                      237.911,  238.723,  Anhydrite
                      238.723,  239.807,  Grey vf-f sandstone
                      239.807,  240.774,  Red siltstone
                      240.774,  241.122,  Dolomite
                      241.122,  241.702,  Grey siltstone
                      241.702,  243.095,  Dolomite
                      243.095,  246.654,  Grey vf-f sandstone
                      246.654,  247.234,  Dolomite
                      247.234,  255.435,  Grey vf-f sandstone
                      255.435,  258.723,  Grey siltstone
                      258.723,  259.729,  Dolomite
                      259.729,  260.967,  Grey siltstone
                      260.967,  261.354,  Dolomite
                      261.354,  267.041,  Grey siltstone
                      267.041,  267.350,  Dolomite
                      267.350,  274.004,  Grey siltstone
                      274.004,  274.313,  Dolomite
                      274.313,  294.816,  Grey siltstone
                      294.816,  295.397,  Dolomite
                      295.397,  296.286,  Limestone
                      296.286,  300.000,  Volcanic
                      """

csv_points = """top, porosity
                1200, 6.4
                1205, 7.3
                1210, 8.2
                1250, 9.2
                1275, 4.3
                1300, 2.2"""

lappy_list = [Interval(**{'top': 50,
                                  'base': 60,
                                  'components': [Component({'lithology': 'dolomite'}),]}),
                      Interval(**{'top': 55,
                                  'base': 75,
                                  'components': [Component({'lithology': 'limestone'}),]}),
                      Interval(**{'top': 75,
                                  'base': 80,
                                  'components': [Component({'lithology': 'volcanic'}),]}),
                      Interval(**{'top': 78,
                                  'base': 90,
                                  'components': [Component({'lithology': 'anhydrite'}),]})
                      ]

top_dict = {'Ardmore': 2510.03,
            'Cody': 2521.61,
            'Sussex Upper Top': 2521.61,
            'Sussex Lower Top': 2527.71,
            'Sussex Lower Base': 2528.62,
            'Sussex Upper Base': 2529.54,
            'Niobrara': 2530.75
           }

data_dict = {'top': [70.273, 70.744],
             'facies': [3, 0],
             'description': ['brown shale with interbedded sandstone pebbles', 'cyrstalline dolomite'],
            }


def test_error():
    """Test the generic error.
    """
    with pytest.raises(StriplogError):
        Striplog([])


def test_striplog():
    """Test most of the things.
    """
    r1 = Component({'lithology': 'sand'})
    r2 = Component({'lithology': 'shale'})
    r3 = Component({'lithology': 'limestone'})

    # Bottom up: elevation order
    iv1 = Interval(120, 100, components=[r1])
    iv2 = Interval(150, 120, components=[r2])
    iv3 = Interval(180, 160, components=[r1, r2])
    iv4 = Interval(200, 180, components=[r3, r2])

    s1 = Striplog([iv1, iv2])
    s2 = Striplog([iv3, iv4])
    s = s1 + s2
    assert s.order == 'elevation'
    assert len(s) == 4
    assert s.start.z == 100
    assert s.stop.z == 200
    assert s.__repr__() != ''
    assert s.__str__() != ''

    s_rev = s.invert(copy=True)
    assert s_rev.order == 'depth'
    x = s.invert()
    assert x is None
    assert s.order == 'depth'
    assert s[0].top.z == 100

    # Top down: depth order
    iv1 = Interval(80, 120, components=[r1])
    iv2 = Interval(120, 150, components=[r2])
    iv3 = Interval(180, 200, components=[r1, r2])
    iv4 = Interval(200, 250, components=[r3, r2])

    s = Striplog([iv1, iv2, iv3, iv4])
    assert s.order == 'depth'
    assert len(s) == 4
    assert s.start.z == 80
    assert s.stop.z == 250
    assert s._Striplog__strict()

    listy = [iv.thickness for iv in s]
    assert len(listy) == 4

    s[2] = Interval(180, 190, components=[r1, r2])
    assert len(s.find_gaps()) == 2

    # Crop.
    x = s.crop((110, 210), copy=True)
    assert x.start.z == 110

    # To csv
    csv = x.to_csv(header=True)
    assert csv[:3] == 'Top'

    # Add.
    assert len(s + iv4) == 5


def test_from_dict():
    """
    Test gen from dictionary.
    """
    striplog = Striplog.from_dict(top_dict)
    assert len(striplog) == 7
    assert striplog.thinnest().summary() == '0.00 m of Cody'


def test_descriptions_and_data():
    """Test that descriptions are parsed into Components when
    miscellaneous data is present.
    """
    strip = Striplog._build_list_of_Intervals(data_dict)
    assert len(strip[0].components[0]) == 2


def test_from_image():
    """Test the generation of a striplog from an image.
    """
    legend = Legend.builtin('NSDOE')
    imgfile = "tests/data/M-MG-70_14.3_135.9.png"
    striplog = Striplog.from_image(imgfile, 200, 300, legend=legend)
    assert len(striplog) == 26
    assert striplog[-1].primary.summary() == 'Volcanic'
    assert np.floor(striplog.find('sandstone').cum) == 15
    assert striplog.read_at(260).primary.lithology == 'siltstone'
    assert striplog.to_las3() != ''
    assert striplog.cum == 100.0
    assert striplog.thickest().primary.lithology == 'anhydrite'
    assert striplog.thickest(n=7)[1].primary.lithology == 'sandstone'
    assert striplog.thinnest().primary.lithology == 'dolomite'
    assert striplog.thinnest(n=7)[1].primary.lithology == 'siltstone'

    # To and from log.
    log, basis, table = striplog.to_log(step=0.1524, return_meta=True)
    assert log[5] == 2.0
    strip = Striplog.from_log(log, basis=basis, components=table)
    assert len(strip) == len(striplog)
    strip2 = Striplog.from_log(log, basis=basis, cutoff=3, legend=legend)
    assert len(strip2) == 18

    # Extract log onto striplog.
    striplog = striplog.extract(log, basis=basis, name="Log", function=np.mean)
    assert striplog[0].data['Log'] == 2.0

    # Indexing.
    indices = [2, 7, 20]
    del striplog[indices]
    assert len(striplog.find_gaps()) == len(indices)

    # Prune and anneal.
    striplog = striplog.prune(limit=1.0, keep_ends=True)
    assert len(striplog) == 14

    striplog = striplog.anneal()
    assert not striplog.find_gaps()  # Should be None

    striplog = striplog.merge_neighbours()
    assert len(striplog) == 11

    rock = striplog.find('sandstone')[1].components[0]
    assert rock in striplog

    # Anneal up or down
    s = striplog[[1, 3]]
    assert s.anneal(mode='up')[1].top.z == s[0].base.z
    assert s.anneal(mode='down')[0].base.z == s[1].top.z


def test_from_descriptions():
    """Test the CSV route.
    """
    lexicon = Lexicon.default()
    strip2 = Striplog.from_descriptions(text=csv_intervals, lexicon=lexicon)
    assert len(strip2.unique) == 7


def test_points():
    """Test a striplog of points.
    """
    points = Striplog.from_csv(text=csv_points, points=True)
    assert len(points) == 6
    assert points.order == 'none'


def test_from_las3():
    """Test the LAS3 route.
    """
    lexicon = Lexicon.default()
    s = Striplog.from_las3(las3, lexicon=lexicon)
    assert len(s) == 14


def test_from_array():
    """Test the array route.
    Deprecated.
    """
    lexicon = Lexicon.default()

    a = [(100, 200, 'red sandstone'),
         (200, 250, 'grey shale'),
         (200, 250, 'red sandstone with shale stringers'),
         ]

    with pytest.warns(DeprecationWarning):
        s = Striplog._from_array(a, lexicon=lexicon)

    assert s.__str__() != ''


def test_striplog_intersect():
    """Test intersection. This example is from the tutorial.
    """
    chrono = Striplog([Interval(**{'top': 0,
                                   'base': 60,
                                   'components': [Component({'age': 'Holocene'})]
                                   }),
                       Interval(**{'top': 60,
                                   'base': 75,
                                   'components': [Component({'age': 'Palaeogene'})]
                                   }),
                       Interval(**{'top': 75,
                                   'base': 100,
                                   'components': [Component({'age': 'Cretaceous'})]
                                   }),
                       ])
    legend = Legend.builtin('NSDOE')
    imgfile = "tests/data/M-MG-70_14.3_135.9.png"
    strip = Striplog.from_image(imgfile, 14.3, 135.9, legend=legend)
    sands = strip.find('sandstone')
    cretaceous = chrono.find('Palaeogene')
    cret_sand = sands.intersect(cretaceous)
    assert len(cret_sand) == 3
    assert cret_sand.stop.z == 75


def test_striplog_merge_overlaps():
    """Test merging. This example is from the tutorial.
    """
    lappy = Striplog(lappy_list)
    assert lappy.find_overlaps(index=True) == [0, 2]
    assert lappy.merge_overlaps() is None
    assert lappy.find_overlaps() is None


def test_striplog_merge():
    """Test new merging. This example is from the tutorial.
    """
    lappy = Striplog(lappy_list)
    assert len(lappy.merge('top')) == 4
    assert lappy.merge('top')[0].base.z == 55
    assert lappy.merge('top', reverse=True)[0].base.z == 60


def test_striplog_union():
    """Test union.
    """
    lappy = Striplog([Interval(**{'top': 0,
                                  'base': 60,
                                  'components': [Component({'lithology': 'dolomite'}),]}),
                      Interval(**{'top': 55,
                                  'base': 75,
                                  'components': [Component({'lithology': 'limestone'}),]}),
                      ])
    lippy = Striplog([Interval(**{'top': 0,
                                  'base': 30,
                                  'components': [Component({'lithology': 'marl'}),]}),
                      ])
    u = lappy.union(lippy)
    assert len(u) == 2
    assert len(u[0].components) == 2


def test_fill():
    gappy = Striplog([Interval(**{'top': 0,
                                  'base': 60,
                                  'components': [Component({'lithology': 'dolomite'}),]}),
                      Interval(**{'top': 65,
                                  'base': 75,
                                  'components': [Component({'lithology': 'limestone'}),]}),
                  ])

    f = gappy.fill(Component({'lithology': 'marl'}))
    assert len(f) == 3
    assert f[1].primary == Component({'lithology': 'marl'})

    f = gappy.fill()
    assert not f[1]


def test_histogram():
    """Test histogram. This example is from the tutorial.
    """
    lexicon = Lexicon.default()
    striplog = Striplog.from_las3(las3, lexicon=lexicon)
    thicks, *_ = striplog.histogram(plot=False)  # See test_plots for plots.
    t = [123.005, 7.919, 5.504, 4.022, 2.964]
    assert np.allclose(t, thicks)


def test_petrel():
    """Test we can load petrel text.
    """
    # What to include: only rows with Well = P-108
    include = {'Well': lambda x: x == 'P-108'}

    # Rename the Surface field as Name.
    remap = {'Surface': 'Name'}

    # What to exclude: any rows with Name = TD
    exclude = {'Name': lambda x: x == 'TD'}

    # What to transform before using.
    function = {'Z': lambda x: -x,
                'Name': lambda x: x.replace('Maguma', 'Meguma')}

    # Which fields to leave out of the result, apart from those that are Null.
    ignore = ['Edited by user',
              'Locked to fault',
              'Used by dep.conv.',
              'Well', 'Symbol']

    # Do the thing!
    s = Striplog.from_petrel("tests/data/petrel.dat",
                             include=include,
                             exclude=exclude,
                             remap=remap,
                             ignore=ignore,
                             function=function,
                             points=False,
                             null=-999.0
                             )

    print(s)

    assert len(s) == 4
    assert s[3].base.z == 1056
    assert s[3].data['Name'] == 'Meguma'



"""
Define a suite a tests for the utils module.
"""
import pytest
import numpy as np

from striplog.utils import null
from striplog.utils import partialmethod
from striplog.utils import rgb_to_hex, hex_to_rgb
from striplog.utils import hex_to_name, name_to_hex
from striplog.utils import hex_is_dark, text_colour_for_hex
from striplog.utils import list_and_add
from striplog.utils import tops_from_loglike
from striplog.utils import binary_dilation, binary_erosion


def test_null():
    """Test the null function on itself.
    """
    assert null(null) == null


def test_binary():
    arr = [1,0,0,0,1,1,1,0,1]
    assert np.all(binary_dilation(arr, 1) == arr)
    assert np.all(binary_dilation(arr, 3) == [1,1,0,1,1,1,1,1,1])
    assert np.all(binary_erosion(arr, 3) == [0,0,0,0,0,1,0,0,0])


def test_partial():
    """Pretty sure this test doesn't test what I need. See coverage report.
    """
    def f(a, b):
        return a + b

    def g(a, b):
        return b

    p = partialmethod(f, b=0)
    q = partialmethod(g, b=None)

    # Can't see how to use utils.py lines 59-61.
    assert p(1)
    assert getattr(q, 'keywords', None)
    assert q(1) is None


def test_colours():
    """Test colour conversions.
    """
    assert rgb_to_hex([0, 0, 0]) == '#000000'
    assert rgb_to_hex([0, 0.5, 0.5]) == '#008080'
    assert rgb_to_hex([255, 128, 128]) == '#ff8080'
    assert hex_to_rgb('#000000') == (0, 0, 0)
    assert hex_to_rgb('#ff8080') == (255, 128, 128)  # case

    # And exceptions:
    with pytest.raises(Exception):
        _ = rgb_to_hex([0, 0, -1])
        assert _

    with pytest.raises(Exception):
        _ = rgb_to_hex([0, 0, 256])
        assert _

    with pytest.raises(Exception):
        _ = rgb_to_hex([0, 0.1, 2])
        assert _


def test_names():
    """Test colour names.
    """
    assert name_to_hex('red') == '#ff0000'
    assert name_to_hex('Red') == '#ff0000'
    assert hex_to_name('#ff0000') == 'red'
    assert hex_to_name('#ff0000') == 'red'
    assert hex_to_name('cerulean') is None


def test_hex_is_dark():
    """Test ability to recover correct colour for text.
    """
    assert not hex_is_dark('#ffff00')
    assert hex_is_dark('#330000')
    assert text_colour_for_hex('#ffff00') == '#000000'
    assert text_colour_for_hex('#121111') == '#ffffff'


def test_list_and_add():
    a = ['this', 'that', 'other']
    b = 'those'
    assert len(list_and_add(b, b)) == 2
    assert len(list_and_add(a, b)) == 4
    assert len(list_and_add(b, a)) == 4


def test_tops_from_loglike():
    a = [1,1,1,2,2,2,-1,-1,-1,np.nan,np.nan,-2,-2,-2]
    tops, values = tops_from_loglike(a)
    assert len(values) == 5

    a = [1,1,1,2,2,2,-1,-1,-1,-2,-2,-2,-2,-2]
    tops, values = tops_from_loglike(a)
    assert len(values) == 4



"""
Define a suite a tests for the canstrat reader.
"""
import pytest

from striplog import Striplog


def test_canstrat():
    """Read a file.
    """
    s = Striplog.from_canstrat('tests/data/canstrat.dat')
    assert len(s) == 28
    assert abs(s[3].data['grains_mm'] - 0.0012) < 0.0000001



"""
Define a suite a tests for the Striplog module.
"""
import numpy as np
import pytest

from striplog.markov import Markov_chain, MarkovError

data = "sssmmmlllmlmlsslsllsmmllllmssssllllssmmlllllssssssmmmmsmllllssslmslmsmmmslsllll"""

def test_basics():
    data = [[0, 37,  3,  2],
            [21,  0, 41, 14],
            [20, 25,  0,  0],
            [1, 14,  1,  0]]

    m = Markov_chain(data, states=['A', 'B', 'C', 'D'])

    assert str(m) == "Markov_chain(179 transitions, states=['A', 'B', 'C', 'D'], step=1, include_self=False)"

    ans = (35.73687369691601, 11.070497693516351, 0.9999989278539752)
    assert np.allclose(m.chi_squared(), ans)

    ans = np.array([[0., 31.27069125, 8.17143874, 2.55787001],
                    [31.28238248, 0., 34.05692583, 10.66069169],
                    [8.17137105, 34.04391563, 0., 2.78471333],
                    [2.5579797, 10.65716447, 2.78485582, 0.]])
    assert np.allclose(m.expected_counts, ans)


def test_sequence():
    m = Markov_chain.from_sequence(data, include_self=True)

    assert len(m._state_counts) == 3

    ans = np.array([[19., 5., 7.],
                    [6., 9., 5.],
                    [7., 6., 14.]])
    assert np.allclose(m.observed_counts, ans)

    ans = np.array([[0.49712747, 0.19796476, 0.30490777],
                    [0.49712747, 0.19796476, 0.30490777],
                    [0.49712747, 0.19796476, 0.30490777]])
    assert np.allclose(m.expected_freqs, ans)

    ans = np.array([[-2.24633883, -2.14054029, -2.81568096],
                    [-1.81677174,  1.82886491, -0.94412655],
                    [-2.68890472, -0.51627836,  0.76836845]])
    assert np.allclose(m.normalized_difference, ans)


def test_generate():
    m = Markov_chain.from_sequence(data, include_self=True)
    
    assert len(m.generate_states()) == 10


def test_step_2():
    m = Markov_chain.from_sequence(data, include_self=True, step=2)
    
    assert m.observed_freqs.ndim == 3

def test_graph_fail():
    m = Markov_chain.from_sequence(data, include_self=True, step=2)
    with pytest.raises(MarkovError):
        G = m.as_graph()

# Graph itself tested via plot in test_plots.py.



"""
Define a suite a tests for the Decor class in the LEgend module.
"""
import pytest

from striplog import Decor
from striplog import Legend
from striplog.legend import LegendError
from striplog import Component

r = {'colour': 'grey',
     'grainsize': 'vf-f',
     'lithology': 'sand'}

r3 = {'colour': 'red',
      'grainsize': 'vf-f',
      'lithology': 'sandstone'}


def test_decor():
    """Test decor basics.
    """
    rock = Component(r)
    rock3 = Component(r3)

    d = Decor({'colour': '#FF0000',
               'component': rock})
    d1 = Decor({'colour': '#F80',
               'component': rock3})
    d2 = Decor({'colour': '(255, 128, 0)',
               'component': rock3})
    d3 = Decor({'colour': 'orange',
               'component': rock3})

    l = d + d3
    assert isinstance(l, Legend)
    assert len(l) == 2
    assert d.rgb == (255, 0, 0)
    assert Decor.random(rock3).colour != ''
    assert d1.colour == '#ff8800'
    assert d2.colour == '#ff8000'
    assert d3.colour == '#ffa500'


def test_decor_html():
    """For jupyter notebook
    """
    r = {'lithology': 'sand'}
    rock = Component(r)
    d = {'color': '#267022', 'component': rock, 'width': 3}
    decor = Decor(d)
    component_row = """<tr><td><strong>component</strong></td><td style="color:black; background-color:white"><table><tr><td><strong>lithology</strong></td><td>sand</td></tr></table></td></tr>"""
    hatch_row = """<tr><td><strong>hatch</strong></td><td style="color:black; background-color:white">None</td></tr>"""
    colour_row = """<tr><td><strong>colour</strong></td><td style="color:#ffffff; background-color:#267022">#267022</td></tr>"""
    width_row = """<tr><td><strong>width</strong></td><td style="color:black; background-color:white">3.0</td></tr>"""
    html = decor._repr_html_()
    assert component_row in html
    assert hatch_row in html
    assert colour_row in html
    assert width_row in html


def test_decor_errors():
    """Test decor errors.
    """
    rock = Component(r)

    # No component
    with pytest.raises(LegendError):
        Decor({'colour': 'red'})

    # No decoration
    with pytest.raises(LegendError):
        Decor({'component': rock})

    # Bad colour
    with pytest.raises(LegendError):
        Decor({'colour': 'blurple',
               'component': rock})



"""
Define a suite a tests for the Legend module.
"""
import pytest

from striplog import Legend
from striplog import Decor
from striplog import Component
from striplog.legend import LegendError

csv_text = u"""colour, width, component lithology, component colour, component grainsize
#F7E9A6, 3, Sandstone, Grey, VF-F
#FF99CC, 2, Anhydrite, ,
#DBD6BC, 3, Heterolithic, Grey,
#FF4C4A, 2, Volcanic, ,
#86F0B6, 5, Conglomerate, ,
#FFD073, 4, Sandstone, Red, C-M
#A6D1FF, 3, Limestone, ,
#FFDBBA, 3, Sandstone, Red, VF-F"""

csv_duplicate = u"""colour, width, component lithology, component colour, component grainsize
#F7E9A6, 3, Sandstone, Grey, VF-F
#FF99CC, 2, Anhydrite, ,
#FF99DD, 3, Anhydrite, ,
#DBD6BC, 3, Heterolithic, Grey,"""

r = {'colour': 'grey',
     'grainsize': 'vf-f',
     'lithology': 'sand'}

r3 = {'colour': 'red',
      'grainsize': 'vf-f',
      'lithology': 'sandstone'}


def test_legend():
    """Test all the basics.
    """
    legend = Legend.from_csv(text=csv_text)
    assert legend[0].colour == '#f7e9a6'
    assert legend.max_width == 5
    assert legend.__str__() != ''
    assert legend.__repr__() != ''

    assert len(legend[[3, 4]]) == 2
    assert len(legend[3:5]) == 2

    rock = Component(r)
    assert legend.get_colour(rock) == '#eeeeee'
    assert rock not in legend

    d = Decor({'colour': 'red',
               'component': rock})
    length = len(legend)
    legend[3] = d
    assert len(legend) == length
    assert legend[3].component == rock
    assert d in legend

    rock3 = Component(r3)
    assert legend.get_colour(rock3) == '#ffdbba'
    assert legend.get_width(rock3) == 3.0

    c = legend.get_component('#f7e9a6')
    assert c.lithology == 'sandstone'
    c2 = legend.get_component('#f7e9a7', tolerance=30)
    assert c2.lithology == 'sandstone'

    colours = [d.colour for d in legend]
    assert len(colours) == 8

    assert Legend.random(rock3)[0].colour != ''

    l = Legend.random([rock, rock3])
    assert len(l) == 2
    assert getattr(l[-1], 'colour') != ''
    assert l.to_csv() != ''
    assert l.max_width == 1.0

    l = Legend.random([rock, rock3], width=True, colour='#abcdef')
    assert getattr(l[0], 'colour') == '#abcdef'

    # Test sums.
    summed = legend + l
    assert len(summed) == 10

    summed_again = legend + d
    assert len(summed_again) == 9

    summed_again_again = d + legend
    assert len(summed_again_again) == 9

    # Test equality.
    assert not d == legend


def test_legend_builtins():
    """Test the builtins.
    """
    assert len(Legend.builtin('nsdoe')) == 18
    assert len(Legend.builtin('nagmdm__6_2')) == 206

    # And builtin timescale.
    assert len(Legend.builtin_timescale('isc')) == 240


def test_tolerance_warning(recwarn):
    """Test warning triggers if tolerance too low.
    """
    legend = Legend.from_csv(text=csv_text)
    legend.get_component('#f7e9a7', tolerance=0)
    w = recwarn.pop()
    assert issubclass(w.category, UserWarning)
    assert 'tolerance of 0' in str(w.message)
    assert w.lineno


def test_duplicate_warning(recwarn):
    """Test warning triggers if duplicate component in CSV.
    """
    Legend.from_csv(text=csv_duplicate)
    w = recwarn.pop()
    assert issubclass(w.category, UserWarning)
    assert 'duplicate' in str(w.message)
    assert w.lineno


def test_error():
    """Test errors are raised.
    """
    rock = Component(r)

    # Adding incompatible things
    legend = Legend.from_csv(text=csv_text)
    with pytest.raises(LegendError):
        _ = legend + rock
        assert _

    # Tolerance not allowed.
    with pytest.raises(LegendError):
        legend.get_component('#f7e9a7', tolerance=-1)



"""
Define a suite a tests for the Interval module.
"""
import pytest

from striplog import Lexicon
from striplog import Interval
from striplog import Component
from striplog.interval import IntervalError

r = {'colour': 'grey',
     'grainsize': 'vf-f',
     'lithology': 'sand'}


def test_error():
    """ Test the IntervalError.
    """
    lexicon = Lexicon.default()
    interval = Interval(20, 40, "Grey sandstone.", lexicon=lexicon)
    with pytest.raises(IntervalError):
        error = interval + 'this will raise'
        assert error


def test_interval():
    """ Test intervals.
    """
    lexicon = Lexicon.default()
    interval = Interval(20, 40, "Grey sandstone.", lexicon=lexicon)
    assert interval.primary.lithology == 'sandstone'
    fmt = "{colour} {lithology}"
    answer = '20.00 m of grey sandstone'
    assert interval.summary(fmt=fmt) == answer

    interval_2 = Interval(40, 65, "Red sandstone.", lexicon=lexicon)
    assert interval_2 != interval
    answer = '20.00 m of grey sandstone'
    # Max gives uppermost
    assert max(interval, interval_2).summary(fmt=fmt) == answer

    iv = interval_2 + interval
    assert len(iv.components) == 2
    assert iv.base.z - iv.top.z == 45.0

    rock = Component(r)
    iv = interval + rock
    assert len(iv.components) == 2


def test_interval_html():
    """For jupyter notebook
    """
    interval = Interval(10, 20, components=[Component(r)])
    html_start = '<table><tr><td style="width:2em; background-color:#DDDDDD" rowspan="6"></td><td><strong>top</strong></td><td>10.0</td></tr><tr><td><strong>primary</strong></td><td><table><tr><td><strong>'
    # Can't test all of it because the component attributes are given in
    # random order.
    assert interval._repr_html_()[:187] == html_start

# Depth ordered
i1 = Interval(top=61, base=62.5, components=[Component({'lithology': 'limestone'})])
i2 = Interval(top=62, base=63, components=[Component({'lithology': 'sandstone'})])
i3 = Interval(top=62.5, base=63.5, components=[Component({'lithology': 'siltstone'})])
i4 = Interval(top=63, base=64, components=[Component({'lithology': 'shale'})])
i5 = Interval(top=63.1, base=63.4, components=[Component({'lithology': 'dolomite'})])

# Elevation ordered
i8 = Interval(top=200, base=100, components=[Component({'lithology': 'sandstone'})])
i7 = Interval(top=150, base=50, components=[Component({'lithology': 'limestone'})])
i6 = Interval(top=100, base=0, components=[Component({'lithology': 'siltstone'})])


def test_interval_invert():
    """Test inverting works.
    """
    i = i1.invert(copy=True)
    assert i.order == 'elevation'
    ii = i2.invert()
    assert ii is None
    assert i2.order == 'elevation'
    i2.invert()


def test_interval_binary_relationships():
    """ Test the binary relationships.
    """
    assert i3 != i2
    assert i2 > i4 > i5
    assert min(i1, i2, i5).summary() == "0.30 m of dolomite"
    iv = i2 + i3
    assert iv.thickness == 1.5

    assert i6.relationship(i7) == 'partially'
    assert i5.relationship(i4) == 'containedby'

    assert i1.partially_overlaps(i2)
    assert i2.partially_overlaps(i3)
    assert not i2.partially_overlaps(i4)
    assert i6.partially_overlaps(i7)
    assert i7.partially_overlaps(i6)
    assert not i6.partially_overlaps(i8)
    assert i5.is_contained_by(i3)
    assert i5.is_contained_by(i4)
    assert not i5.is_contained_by(i2)


def test_interval_binary_operations():
    """ Test the binary operations.
    """
    assert '40.0%' in i1.union(i3).description
    assert len(i3.difference(i5)) == 2
    assert i1.intersect(i2, blend=False).description == ''

    x = i4.merge(i5)
    x[-1].base = 65
    assert len(x) == 3
    assert x.stop.z == 65.0


def test_interval_binary_errors():
    """Test errors thrown by binary operations.
    """
    with pytest.raises(IntervalError):
        i1.merge(i8)
    with pytest.raises(IntervalError):
        i1.merge(i4)
    with pytest.raises(IntervalError):
        i1.intersect(i4)



"""
Define a suite a tests for the Lexicon module.
"""
from striplog import Lexicon


def test_lexicon():
    """All the tests...
    """
    lexicon = Lexicon.default()
    s = lexicon.__str__()
    assert s is not ''
    assert lexicon.__repr__() is not ''
    assert lexicon.find_synonym('Halite') == 'salt'
    assert len(lexicon.categories) == 5

    s = "lt gn ss w/ sp gy sh"
    answer = 'lighter green sandstone with spotty gray shale'
    assert lexicon.expand_abbreviations(s) == answer

    fname = "tests/data/lexicon.json"
    l = Lexicon.from_json_file(fname)
    assert l.__repr__() is not ''



"""
Define a suite a tests for the Position module.
"""
import pytest

from striplog import Position
from striplog.position import PositionError


def test_incomplete_error():
    """ Test the PositionError.
    """
    with pytest.raises(PositionError):
        Position(middle=None, upper=1)


def test_incomplete_coords_error():
    """ Test the PositionError.
    """
    with pytest.raises(PositionError):
        Position(middle=1, x=1000)


def test_position():
    """ Test intervals.
    """
    meta = {'character': 'erosive',
            'rugosity': '3 m',
            }
    position = Position(middle=None,
                        upper=1,
                        lower=2,
                        x=1000,
                        y=2000,
                        meta=meta)
    assert position.z == 1.5
    assert position.span == (2, 1)
    assert position.uncertainty == 1.0
    position.invert()
    assert position.span == (1, 2)


def test_position_html():
    """For jupyter notebook
    """
    position = Position(**{'upper': 75, 'lower': 85})
    html = """<table><tr><td><strong>upper</strong></td><td>75.0</td></tr><tr><td><strong>middle</strong></td><td></td></tr><tr><td><strong>lower</strong></td><td>85.0</td></tr></table>"""
    assert position._repr_html_() == html




import sys
import glob
import re


def simplify_credits(html):
    """
    Replace the credit part of the HTML footer. Return the new text.
    """
    s = r"Created using <a href=\"https://www\.sphinx.+?Furo theme</a>."
    pattern = re.compile(s, flags=re.DOTALL)

    new_s = '<a href="https://creativecommons.org/licenses/by/4.0/">CC BY 4.0</a>'
    new_s += ' | Created using Sphinx & Furo'

    return pattern.sub(new_s, html)

def main(path):
    """
    Process the HTML files in path, save in place (side-effect).
    """
    fnames = glob.glob(path.strip('/') + '/*.html')
    for fname in fnames:
        with open(fname, 'r+') as f:
            html = f.read()

            new_html = simplify_credits(html)

            f.seek(0)
            f.write(new_html)
            f.truncate()
    return


if __name__ == '__main__':
    _ = main(sys.argv[1])



"""
Defines a suite a tests for images run with:

    py.test --mpl

To generate new test images, see the instructions in
striplog/run_tests.py

https://pypi.python.org/pypi/pytest-mpl/0.3
"""
import pytest

from striplog import logo


params = {'tolerance': 20,
          'savefig_kwargs': {'dpi': 100},
          }


@pytest.mark.mpl_image_compare(**params)
def test_striplog_logo():
    """
    Tests mpl image of striplog logo.
    """
    fig = logo.plot()
    return fig



import pathlib
import os

import matplotlib.font_manager as fm
import matplotlib.pyplot as plt
import numpy as np

from .striplog import Striplog, Legend


def plot(fname=None):
    """
    Plots the Striplog logo.
    Requires Snare Drum font, by Nate Halley.

    Args:
        fname (str): Filename to save the plot to.

    Returns:
        matplotlib.figure.Figure: The figure object.
    """
    csv_string = """top, base, comp lithology
                    0,   5, Shale
                    5,   9, Limestone
                    9,  28, Conglomerate
                    28, 32, Sandstone
                    32, 36, Conglomerate
                    36, 51, Sandstone
                    51, 57, Siltstone
                    57, 65, Shale
                    65, 68, Mudstone
                    68, 71, Shale
                    71, 83, Mudstone
                    83, 95, Shale
                    95,100, Mudstone"""
    strip = Striplog.from_csv(text=csv_string)

    l = u"""colour, width, component lithology
            #feec97, 6, Sandstone
            #fdd218, 7, Conglomerate
            #c6b259, 5, Siltstone
            #3ab4ff, 5, Limestone
            #d2d2d2, 5, Mudstone
            #909090, 4, Shale"""
    legend = Legend.from_csv(text=l)

    # Read the file. NB This is not ZIP-safe.
    here = pathlib.Path(os.path.dirname(os.path.abspath(__file__)))
    im = plt.imread(here / 'striplog_logotype.png')
    im[im==0] = np.nan

    # Make the plot.
    fig, ax = plt.subplots(figsize=(5, 5))

    strip.plot(ax=ax, legend=legend, aspect=1)
    ax.axis('off')

    logo_ax = fig.add_axes([0.162, 0.585, 0.7, 0.2], anchor='NE', zorder=1, facecolor='none')
    logo_ax.imshow(im, cmap='gray', interpolation='none')
    logo_ax.axis('off')

    if fname is not None:
        fig.savefig(fname, dpi=200)

    plt.close()

    return fig
